import java.util.Arrays;

public class Nodo {
	private int tamanhoTabuleiro = 0;

	// tabuleiro
	private int m[][] = null;

	private Nodo pai = null;

	// armazena posicao do espao vazio no tabuleiro para facilitar os
	// calculos...
	// a posicao no tabuleiro  dada por
	// 1 2 3
	// 4 5 6
	// 7 8 9
	private int posBranco;

	// Construtores
	public Nodo(int tamanhoTab) {
		this.tamanhoTabuleiro = tamanhoTab;
		m = new int[this.tamanhoTabuleiro][this.tamanhoTabuleiro];
		int cont = 1;
		for (int i = 0; i < this.tamanhoTabuleiro; i++)
			for (int j = 0; j < this.tamanhoTabuleiro; j++)
				this.m[i][j] = cont++;
		this.m[this.tamanhoTabuleiro - 1][this.tamanhoTabuleiro - 1] = 0;
		this.posBranco = this.tamanhoTabuleiro * this.tamanhoTabuleiro;
		this.pai = null;
	}

	public Nodo(int _m[][], int _posBranco, Nodo _pai) {
		this(_m.length);
		this.m = _m;
		this.posBranco = _posBranco;
		this.pai = _pai;
	}

	// Mtodos...
	public int getTamTabuleiro() {
		return this.tamanhoTabuleiro;
	}

	public boolean equals(Nodo x) {
		for (int i = 0; i < this.tamanhoTabuleiro; i++)
			for (int j = 0; j < this.tamanhoTabuleiro; j++)
				if (this.m[i][j] != x.getPecaNoTabuleiro(i, j))
					return false;
		return true;
	}

	public int getPosBranco() {
		return this.posBranco;
	}

	public Nodo getPai() {
		return this.pai;
	}

	public int[][] getTabuleiro() {
		return this.m;
	}

	public int getPecaNoTabuleiro(int x, int y) {
		return this.m[x][y];
	}

	/**
	 * faz a movimentacao do tabuleiro verificando se o movimento  vlido. as
	 * posicoes do tabuleiro vao de 1 a
	 * (this.tamanhaTabuleiro*this.tamanhaTabuleiro) a posicao de origem 
	 * sempre o valor do vazio
	 * 
	 */
	public boolean validaMovimento(int destino) {
		try {
			if (posBranco < 1
					|| posBranco > (this.tamanhoTabuleiro * this.tamanhoTabuleiro))
				throw new Exception("Origem invlida");
			if (destino < 1
					|| destino > (this.tamanhoTabuleiro * this.tamanhoTabuleiro))
				throw new Exception("Destino invlido");
			int yOrigem = (posBranco - 1) % this.tamanhoTabuleiro;// Captura
			// coordenada
			// y (de 0 a
			// 2)
			int xOrigem = Math.abs((posBranco - 1) / this.tamanhoTabuleiro);// Captura
			// coordenada
			// x
			// (de
			// 0 a
			// 2)
			int yDestino = (destino - 1) % this.tamanhoTabuleiro;
			int xDestino = Math.abs((destino - 1) / this.tamanhoTabuleiro);
			if (this.m[xOrigem][yOrigem] != 0) {
				throw new Exception("Erro Interno: Origem nao tem espaco vazio");
			}
			if ((Math.abs(xOrigem - xDestino) == 1 && yOrigem == yDestino)
					|| (Math.abs(yOrigem - yDestino) == 1 && xOrigem == xDestino))
				return true;
			else
				return false;
		} catch (Exception ex) {
			ex.printStackTrace();
			return false;
		}
	}

	public boolean mover(int destino) {
		if (validaMovimento(destino)) {
			// Faz a troca porque o movimento  valido...
			int yOrigem = (posBranco - 1) % this.tamanhoTabuleiro;// Captura
			// coordenada
			// y (de 0 a
			// 2)
			int xOrigem = Math.abs((posBranco - 1) / this.tamanhoTabuleiro);// Captura
			// coordenada
			// x
			// (de
			// 0 a
			// 2)
			int yDestino = (destino - 1) % this.tamanhoTabuleiro;
			int xDestino = Math.abs((destino - 1) / this.tamanhoTabuleiro);
			int temp = this.m[xOrigem][yOrigem];
			this.m[xOrigem][yOrigem] = this.m[xDestino][yDestino];
			this.m[xDestino][yDestino] = temp;
			this.posBranco = destino;
			return true;
		} else {
			return false;
		}
	}

	public int getQtdeFilhos() {
		int qtdeMovimentosValidos = 0;
		for (int i = 1; i <= (this.tamanhoTabuleiro * this.tamanhoTabuleiro); i++)
			if (this.validaMovimento(i))
				qtdeMovimentosValidos++;
		return qtdeMovimentosValidos;
	}

	public Nodo[] getFilhos() {
		Nodo colecao[] = new Nodo[this.getQtdeFilhos()];
		int index = 0;
		try {
			for (int i = 1; i <= (this.tamanhoTabuleiro * this.tamanhoTabuleiro); i++) {
				if (this.validaMovimento(i)) {
					// cria novo nodo porque metodo clone nao funciona...
					// Nodo nodoTemp = (Nodo)(this.clone()); nao funciona...
					// Foi necessario copiar a matriz a mao...
					int matrizTemp[][] = new int[this.tamanhoTabuleiro][this.tamanhoTabuleiro];
					for (int x = 0; x < this.tamanhoTabuleiro; x++)
						for (int j = 0; j < this.tamanhoTabuleiro; j++)
							matrizTemp[x][j] = this.m[x][j];
					Nodo nodoTemp = new Nodo(matrizTemp, this.posBranco, this);
					nodoTemp.mover(i);
					colecao[index++] = nodoTemp;
				}
			}
			return colecao;
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}

	public Nodo copiarSemPai() {
		int matrizTemp[][] = new int[this.tamanhoTabuleiro][this.tamanhoTabuleiro];
		for (int x = 0; x < this.tamanhoTabuleiro; x++)
			for (int j = 0; j < this.tamanhoTabuleiro; j++)
				matrizTemp[x][j] = this.m[x][j];
		Nodo nodoTemp = new Nodo(matrizTemp, this.posBranco, null);
		return nodoTemp;
	}

	public int[] getMovimentosValidos() {
		int temp[] = new int[this.getQtdeFilhos()];
		int index = 0;
		for (int i = 1; i <= (this.tamanhoTabuleiro * this.tamanhoTabuleiro); i++) {
			if (this.validaMovimento(i)) {
				temp[index] = i;
				index++;
			}
		}
		return temp;
	}

	public boolean isResolvido() {
		for (int i = 0; i < this.tamanhoTabuleiro; i++)
			for (int j = 0; j < this.tamanhoTabuleiro; j++) {
				if ((i == (this.tamanhoTabuleiro-1))
						&& (j == (this.tamanhoTabuleiro-1))) {// se for ultima
					// posicao.
					if (this.m[i][j] != 0)
						return false;
				} else if (this.m[i][j] != ((i * this.tamanhoTabuleiro) + j + 1))
					return false;
			}
		return true;
	}

	public int getHeuristicaNumPecasForaLugar() {
		int soma = 0;
		if (this.m[0][0] != 1)
			soma++;
		if (this.m[0][1] != 2)
			soma++;
		if (this.m[0][2] != 3)
			soma++;
		if (this.m[1][0] != 4)
			soma++;
		if (this.m[1][1] != 5)
			soma++;
		if (this.m[1][2] != 6)
			soma++;
		if (this.m[2][0] != 7)
			soma++;
		if (this.m[2][1] != 8)
			soma++;
		if (this.m[2][2] != 0)
			soma++;
		return soma;
	}

	public int getHeuristicaDistManhattan() {
		int soma = 0;
		int x = 0;
		int y = 0;
		for (int i = 0; i < this.tamanhoTabuleiro; i++) {
			for (int j = 0; j < this.tamanhoTabuleiro; j++) {
				if (this.m[i][j] == 0) {
					// O zero nao pode contar na distancia manhattan...
				} else {
					x = Math.abs((this.m[i][j] - 1) / this.tamanhoTabuleiro);
					y = (this.m[i][j] - 1) % this.tamanhoTabuleiro;
					soma += (Math.abs(x - i) + Math.abs(y - j));
				}
			}
		}
		return soma;
	}

	public int getHeuristicaDistManhattanComVerificacaoDeCantos() {
		int soma = 0;
		int x = 0;
		int y = 0;
		// calculo da distancia manhatan
		for (int i = 0; i < this.tamanhoTabuleiro; i++) {
			for (int j = 0; j < this.tamanhoTabuleiro; j++) {
				if (this.m[i][j] == 0) {
					// O zero nao pode contar na distancia manhattan...
				} else {
					x = Math.abs((this.m[i][j] - 1) / this.tamanhoTabuleiro);
					y = (this.m[i][j] - 1) % this.tamanhoTabuleiro;
					soma += (Math.abs(x - i) + Math.abs(y - j));
				}
			}
		}
		// penalizao dos cantos...
		if (this.m[0][0] != ((0 * this.tamanhoTabuleiro) + 0 + 1))
			soma++;
		if (this.m[0][this.tamanhoTabuleiro-1] != ((0 * this.tamanhoTabuleiro) + this.tamanhoTabuleiro + 1))
			soma++;
		if (this.m[this.tamanhoTabuleiro-1][0] != ((this.tamanhoTabuleiro * this.tamanhoTabuleiro) + 0 + 1))
			soma++;
		if (this.m[this.tamanhoTabuleiro-1][this.tamanhoTabuleiro-1] != 0)
			soma++;
		return soma;
	}

}
